@@ -2,7 +2,8 @@ import logging
 import sys
 import traceback
 from argparse import ArgumentParser, ArgumentTypeError, Namespace
-from typing import Iterator, Optional, Tuple, Union
+from itertools import chain
+from typing import Iterator, List, Optional, Tuple, Union
 
 from dql import __version__, utils
 from dql.catalog import Catalog
@@ -46,7 +47,9 @@ def add_sources_arg(
 def get_parser() -> ArgumentParser:
     parser = ArgumentParser(description="DQL: Data Query Language", prog="dql")
 
-    parser.add_argument("--version", action="version", version=__version__)
+    parser.add_argument(
+        "-V", "--version", action="version", version=__version__
+    )
 
     parent_parser = ArgumentParser(add_help=False)
     parent_parser.add_argument(
@@ -131,6 +134,12 @@ def get_parser() -> ArgumentParser:
         action="store_true",
         help="Do not expand globs (such as * or ?)",
     )
+    parse_cp.add_argument(
+        "--no-dql-file",
+        default=False,
+        action="store_true",
+        help="Do not write a .dql file",
+    )
 
     parse_ls = subp.add_parser(
         "ls", parents=[parent_parser], help="List storage contents"
@@ -213,17 +222,40 @@ def ls_urls(
     long: bool = False,
     *,
     client_config=None,
+    catalog: Optional[Catalog] = None,
     **kwargs,
 ) -> Iterator[Tuple[str, Iterator[str]]]:
     if client_config is None:
         client_config = {}
-    data_storage = DefaultDataStorage()
-    catalog = Catalog(data_storage)
+    if catalog is None:
+        catalog = Catalog(DefaultDataStorage())
+
+    curr_dir = None
+    value_iterables = []
+    for next_dir, values in _ls_urls_flat(
+        sources, long, client_config, catalog, **kwargs
+    ):
+        if curr_dir is None or next_dir == curr_dir:  # type: ignore[unreachable] # noqa: E501
+            value_iterables.append(values)
+        else:
+            yield curr_dir, chain(*value_iterables)  # type: ignore[unreachable] # noqa: E501
+            value_iterables = [values]
+        curr_dir = next_dir
+    if curr_dir is not None:
+        yield curr_dir, chain(*value_iterables)
+
+
+def _ls_urls_flat(
+    sources,
+    long: bool,
+    client_config,
+    catalog: Catalog,
+    **kwargs,
+) -> Iterator[Tuple[str, Iterator[str]]]:
     for source in sources:
-        if Client.get_implementation(source).is_root_url(source):
-            buckets = Client.get_implementation(source).ls_buckets(
-                **client_config
-            )
+        client_cls = Client.get_implementation(source)
+        if client_cls.is_root_url(source):
+            buckets = client_cls.ls_buckets(**client_config)
             if long:
                 values = (
                     long_line_str(b.name, b.created, "") for b in buckets
@@ -239,11 +271,15 @@ def ls_urls(
                     values = (n.long_line_str() for n in nodes)
                 else:
                     values = (n.name_with_dir_ending for n in nodes)
-                yield data_source.node.path, values
+                yield "/".join(data_source.node.path), values
 
 
-def ls_indexed_storages(long: bool = False) -> Iterator[str]:
-    storages = Catalog(DefaultDataStorage()).ls_storages()
+def ls_indexed_storages(
+    long: bool = False, catalog: Optional[Catalog] = None
+) -> Iterator[str]:
+    if catalog is None:
+        catalog = Catalog(DefaultDataStorage())
+    storages = catalog.ls_storages()
     if long:
         for s in storages:
             # TODO: add Storage.created so it can be used here
@@ -253,19 +289,34 @@ def ls_indexed_storages(long: bool = False) -> Iterator[str]:
             yield s.uri
 
 
-def ls(sources, long: bool = False, **kwargs):
+def ls(
+    sources, long: bool = False, catalog: Optional[Catalog] = None, **kwargs
+):
+    if catalog is None:
+        catalog = Catalog(DefaultDataStorage())
     if sources:
         if len(sources) == 1:
-            for _, entries in ls_urls(sources, long=long, **kwargs):
+            for _, entries in ls_urls(
+                sources, long=long, catalog=catalog, **kwargs
+            ):
                 for entry in entries:
                     print(entry)
         else:
-            for source, entries in ls_urls(sources, long=long, **kwargs):
-                print(f"{source}:")
+            first = True
+            for source, entries in ls_urls(
+                sources, long=long, catalog=catalog, **kwargs
+            ):
+                # print a newline between directory listings
+                if first:
+                    first = False
+                else:
+                    print()
+                if source:
+                    print(f"{source}:")
                 for entry in entries:
                     print(entry)
     else:
-        for entry in ls_indexed_storages(long=long):
+        for entry in ls_indexed_storages(long=long, catalog=catalog):
             print(entry)
 
 
@@ -282,20 +333,24 @@ def find(sources, **kwargs):
     yield from catalog.find(sources, **kwargs)
 
 
-def index(sources, **kwargs):
-    data_storage = DefaultDataStorage()
-    catalog = Catalog(data_storage)
+def index(sources, catalog: Optional[Catalog] = None, **kwargs):
+    if catalog is None:
+        catalog = Catalog(DefaultDataStorage())
     catalog.index(sources, **kwargs)
 
 
-def main():  # noqa: C901
+def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
     parser = get_parser()
-    args = parser.parse_args()
+    args = parser.parse_args(argv)
 
     logger.addHandler(logging.StreamHandler())
     logging_level = get_logging_level(args)
     logger.setLevel(logging_level)
 
+    if args.command is None:
+        parser.print_help()
+        return 1
+
     client_config = {
         "aws_endpoint_url": args.aws_endpoint_url,
         "aws_anon": args.aws_anon,
@@ -359,7 +414,7 @@ def main():  # noqa: C901
                 update=bool(args.update),
                 client_config=client_config,
             )
-
+        return 0
     except Exception as exc:  # pylint: disable=broad-except
         print("Error:", exc, file=sys.stderr)
         if logging_level <= logging.DEBUG:
@@ -369,4 +424,4 @@ def main():  # noqa: C901
                 exc.__traceback__,
                 file=sys.stderr,
             )
-        sys.exit(1)
+        return 1
