@@ -8,12 +8,20 @@ from dql.node import Node
 from dql.utils import time_to_str
 
 
+class Status:
+    CREATED = 1
+    PENDING = 2
+    FAILED = 3
+    COMPLETE = 4
+
+
 class StorageRecord(NamedTuple):
     id: str  # pylint:disable=redefined-builtin
     type: str  # pylint:disable=redefined-builtin
     name: str
     timestamp: Optional[datetime]
     expires: Optional[datetime]
+    status: int = Status.CREATED
 
 
 class Storage(StorageRecord):
